import asyncio
from typing import Any, List, Optional, Union
from payments_py.data_models import AgentExecutionStatus, ServiceTokenResultDto
from payments_py.nvm_backend import BackendApiOptions, NVMBackendApi

# Define API Endpoints
SEARCH_TASKS_ENDPOINT = '/api/v1/agents/search/tasks'
SEARCH_STEPS_ENDPOINT = '/api/v1/agents/search/steps'
CREATE_STEPS_ENDPOINT = '/api/v1/agents/{did}/tasks/{taskId}/steps'
UPDATE_STEP_ENDPOINT = '/api/v1/agents/{did}/tasks/{taskId}/step/{stepId}'
GET_AGENTS_ENDPOINT = '/api/v1/agents'
GET_BUILDER_STEPS_ENDPOINT = '/api/v1/agents/steps'
GET_TASK_STEPS_ENDPOINT = '/api/v1/agents/{did}/tasks/{taskId}/steps'
TASK_ENDPOINT = '/api/v1/agents/{did}/tasks'
GET_TASK_ENDPOINT = '/api/v1/agents/{did}/tasks/{taskId}'


class AIQueryApi(NVMBackendApi):
    """
    Represents the AI Query API.

    Args:
        opts (BackendApiOptions): The backend API options

    Methods:
        create_task: Creates a task for an agent to execute
        create_steps: Creates steps for a task
        update_step: Updates a step
        search_tasks: Searches for tasks
        get_task_with_steps: Gets a task with its steps
        get_steps_from_task: Gets the steps from a task
        get_steps: Gets the steps
        get_tasks_from_agents: Gets the tasks from the agents
        search_step: Searches for steps
        get_step: Gets the details of a step
    """
    def __init__(self, opts: BackendApiOptions):
        super().__init__(opts)
        self.opts = opts

    async def subscribe(self, callback: Any, join_account_room: bool = True, join_agent_rooms: Optional[Union[str, List[str]]] = None, subscribe_event_types: Optional[List[str]] = None, get_pending_events_on_subscribe: bool = True):
        """
        It subscribes to the Nevermined network to retrieve new AI Tasks requested by other users.
        This method is used by AI agents to subscribe and receive new AI Tasks sent by other subscribers.

        Args:
            callback (Any): The callback function to be called when a new event is received.
            join_account_room (bool): If True, it will join the account room.
            join_agent_rooms (Optional[Union[str, List[str]]]): The agent rooms to join.
            subscribe_event_types (Optional[List[str]]): The event types to subscribe to.
            get_pending_events_on_subscribe (bool): If True, it will get the pending events on subscribe.
        """
        self.set_subscriber(callback=callback, join_account_room=join_account_room, join_agent_rooms=join_agent_rooms, subscribe_event_types=subscribe_event_types, get_pending_events_on_subscribe=get_pending_events_on_subscribe)
        await self.connect_socket()        
        await asyncio.Event().wait()

    def create_task(self, did: str, task: Any):
        """
        Subscribers can create an AI Task for an Agent. The task must contain the input query that will be used by the AI Agent.
        This method is used by subscribers of a Payment Plan required to access a specific AI Agent or Service. Users who are not subscribers won't be able to create AI Tasks for that Agent.
        Because only subscribers can create AI Tasks, the method requires the access token to interact with the AI Agent/Service.
        This is given using the `queryOpts` object (accessToken attribute).
        
        Args:
            did (str): The DID of the service.
            task (Any): The task to create.

        Example:
            task = {
                "query": "https://www.youtube.com/watch?v=0tZFQs7qBfQ",
                "name": "transcribe",
                "additional_params": [],
                "artifacts": []
            }
            task = subscriber.ai_protocol.create_task(agent.did, task)
            print('Task created:', task.json())
        """
        endpoint = self.parse_url_to_proxy(TASK_ENDPOINT).replace('{did}', did)
        token = self.get_service_token(did)
        return self.post(endpoint, task, headers={'Authorization': f'Bearer {token.accessToken}'})

    def create_steps(self, did: str, task_id: str, steps: Any):
        """
        It creates the step/s required to complete an AI Task.
        This method is used by the AI Agent to create the steps required to complete the AI Task.
        
        Args:
        
            did (str): The DID of the service.
            task_id (str): The task ID.
            steps (Any): The steps to create.
        """
        endpoint = self.parse_url_to_backend(CREATE_STEPS_ENDPOINT).replace('{did}', did).replace('{taskId}', task_id)
        return self.post(endpoint, steps)

    def update_step(self, did: str, task_id: str, step_id: str, step: Any):
        """
        It updates the step with the new information.
        This method is used by the AI Agent to update the status and output of an step. This method can not be called by a subscriber.

        Args:
            did (str): The DID of the service.
            task_id (str): The task ID.
            step_id (str): The step ID.
            step (Any): The step object to update. https://docs.nevermined.io/docs/protocol/query-protocol#steps-attributes
        """
        endpoint = self.parse_url_to_backend(UPDATE_STEP_ENDPOINT).replace('{did}', did).replace('{taskId}', task_id).replace('{stepId}', step_id)
        try:
            return self.put(endpoint, step)
        except Exception as e:
            print('update_step::', e)
            return None

    def search_tasks(self, search_params: Any):
        """
        It searches tasks based on the search parameters associated to the user.

        Args:
            search_params (Any): The search parameters.
        """    
        return self.post(self.parse_url_to_backend(SEARCH_TASKS_ENDPOINT), search_params)

    def get_task_with_steps(self, did: str, task_id: str):
        """
        It returns the full task and the steps resulted of the execution of the task.

        This method is used by subscribers of a Payment Plan required to access a specific AI Agent or Service. Users who are not subscribers won't be able to create AI Tasks for that Agent.


        Args:
            did (str): The DID of the service.
            task_id (str): The task ID.
        """
        endpoint = self.parse_url_to_proxy(GET_TASK_ENDPOINT).replace('{did}', did).replace('{taskId}', task_id)
        token = self.get_service_token(did)
        return self.get(endpoint, headers={'Authorization': f'Bearer {token.accessToken}'})


    def get_steps_from_task(self, did: str, task_id: str, status: Optional[str] = None):
        """
        It retrieves all the steps that the agent needs to execute to complete a specific task associated to the user.
        This method is used by the AI Agent to retrieve information about the tasks created by users to the agents owned by the user.

        Args:
            did (str): The DID of the service.
            task_id (str): The task ID.
            status (Optional[str]): The status of the steps.
        """
        endpoint = self.parse_url_to_backend(GET_TASK_STEPS_ENDPOINT).replace('{did}', did).replace('{taskId}', task_id)
        if status:
            endpoint += f'?status={status}'
        return self.get(endpoint)
    
    def search_step(self, search_params: Any):
        """
        It search steps based on the search parameters. The steps belongs to the tasks part of the AI Agents owned by the user.
        This method is used by the AI Agent to retrieve information about the steps part of tasks created by users to the agents owned by the user.

        Args:
            search_params (Any): The search parameters.
        """
        return self.post(self.parse_url_to_backend(SEARCH_STEPS_ENDPOINT), search_params)

    def get_step(self,  step_id: str):
        """
        Get the details of a step.

        Args:
            did (str): The DID of the service.
            task_id (str): The task ID.
            step_id (str): The step ID.
        """
        result = self.search_step({"step_id": step_id})
        return result.json()['steps'][0]

    def get_steps(self,
                        status: AgentExecutionStatus = AgentExecutionStatus.Pending,
                        dids: List[str] = []):
        """
        It retrieves all the steps that the agent needs to execute to complete the different tasks assigned.
        This method is used by the AI Agent to retrieve information about the steps part of tasks created by users to the agents owned by the user.

        Args:
            status (AgentExecutionStatus): The status of the steps.
            dids (List[str]): The list of DIDs.
        """
        endpoint = f'{self.parse_url_to_backend(GET_BUILDER_STEPS_ENDPOINT)}?'
        if status:
            endpoint += f'&status={status.value}'
        if dids:
            endpoint += f'&dids={",".join(dids)}'
        return self.get(endpoint)

    def get_tasks_from_agents(self):
        """
        It retrieves all the tasks that the agent needs to execute to complete the different tasks assigned.
        This method is used by the AI Agent to retrieve information about the tasks created by users to the agents owned by the user

        """
        return self.get(self.parse_url(GET_AGENTS_ENDPOINT))
    




import uuid

def snake_to_camel(name):
    """
    Convert snake_case to camelCase.
    
    :param name: str
    :return: str
    
    """
    components = name.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])

def generate_step_id():
    """
    Generate a random ID.
    
    :return: str
    
    """
    return f"step-${str(uuid.uuid4())}"

def is_id_valid(id):
    """
    Check if the ID is valid.
    
    :param id: str
    :return: bool

    """
    if not id.startswith('step-'):
        return False 
    else:
        try:
            uuid.UUID(id)
            return True
        except ValueError:
            return False
        
def get_query_protocol_endpoints(server_host: str):
    """
    Get the query protocol endpoints.
    
    :return: dict
    
    """
    return [
        {'POST': f'{server_host}/api/v1/agents/(.*)/tasks'},
        {'GET': f'{server_host}/api/v1/agents/(.*)/tasks/(.*)'},
    ]

def get_ai_hub_open_api_url(server_host: str):
    """
    Get the AI Hub Open API URL.
    
    :return: str
    
    """
    return f'{server_host}/api/v1/rest/docs-json'


import json
import requests
import socketio
import jwt
from typing import Optional, Dict, List, Any, Union

from payments_py.data_models import AgentExecutionStatus, ServiceTokenResultDto
from payments_py.environments import Environment

sio = socketio.AsyncClient(logger=True, engineio_logger=True)


class BackendApiOptions:
    """
    Represents the backend API options.

    Args:
        environment (Environment): The environment.
        api_key (Optional[str]): The Nevermined API Key. This key identify your user and is required to interact with the Nevermined API. You can get your API key by logging in to the Nevermined App. See https://docs.nevermined.app/docs/tutorials/integration/nvm-api-keys
        headers (Optional[Dict[str, str]]): Additional headers to send with the requests
        web_socket_options (Optional[Dict[str, Any]]): Configuration of the websocket connection
    """
    def __init__(self,
                 environment: Environment,
                 api_key: Optional[str] = None,
                 headers: Optional[Dict[str, str]] = None,
                 web_socket_options: Optional[Dict[str, Any]] = None):
        self.api_key = api_key
        self.backend_host = environment.value['backend']
        self.web_socket_host = environment.value['websocket']
        self.proxy_host = environment.value['proxy']
        self.headers = headers or {}
        self.web_socket_options = web_socket_options or {}


class NVMBackendApi:
    def __init__(self, opts: BackendApiOptions):
        self.opts = opts
        self.socket_client = sio
        self.user_room_id = None
        self.has_key = False
        self.callback = None
        self.join_account_room = None
        self.join_agent_rooms = None
        self.subscribe_event_types = None
        self.get_pending_events_on_subscribe = None
        
        default_headers = {
            'Accept': 'application/json',
            **(opts.headers or {}),
            **({'Authorization': f'Bearer {opts.api_key}'} if opts.api_key else {})
        }

        if opts.web_socket_options and opts.web_socket_options.get('bearer_token'):
            opts.web_socket_options['transport_options'] = {
                'websocket': {
                    'extraHeaders': {'Authorization': f'Bearer {opts.web_socket_options["bearer_token"]}'}
                }
            }

        self.opts.headers = default_headers
        self.opts.web_socket_options = {
            **(opts.web_socket_options or {})
        }

        try:
            if self.opts.api_key and len(self.opts.api_key) > 0:
                decoded_jwt = jwt.decode(self.opts.api_key, options={"verify_signature": False})
                client_id = decoded_jwt.get('sub')
                
                # Check if the client_id exists and does not match the specified pattern
                if client_id:# and not re.match(r'^0x[a-fA-F0-9]{40}$', client_id):
                    self.user_room_id = f"room:{client_id}"
                    self.has_key = True
        except Exception:
            self.has_key = False
            self.user_room_id = None 

        try:
            backend_url = self.opts.backend_host.rstrip('/')
            self.opts.backend_host = backend_url
        except Exception as error:
            raise ValueError(f"Invalid URL: {self.opts.backend_host} - {str(error)}")


    def set_subscriber(self, callback, join_account_room, join_agent_rooms, subscribe_event_types, get_pending_events_on_subscribe):
        self.callback = callback
        self.join_account_room = join_account_room
        self.join_agent_rooms = join_agent_rooms
        self.subscribe_event_types = subscribe_event_types
        self.get_pending_events_on_subscribe = get_pending_events_on_subscribe
        self.socket_client.on('_connected', self.connect_handler)

    async def connect_socket(self):
        if not self.has_key:
            raise ValueError('Unable to subscribe to the server because a key was not provided')

        if self.socket_client and self.socket_client.connected:
            return
        
        try:
            print(f"nvm-backend:: Connecting to websocket server: {self.opts.web_socket_host}")
            await self.socket_client.connect(self.opts.web_socket_host, headers=self.opts.headers, transports=["websocket"])
            for i in range(5):
                await self.socket_client.sleep(1)  
                if self.socket_client.connected:
                    break
            print(f"nvm-backend:: Connected: {self.socket_client.connected}")
        except Exception as error:
            raise ConnectionError(f"Unable to initialize websocket client: {self.opts.web_socket_host} - {str(error)}")

    async def disconnect_socket(self):
        if self.socket_client and self.socket_client.connected:
            self.socket_client.disconnect()

    async def connect_handler(self, data):
        await self._subscribe()
        if self.get_pending_events_on_subscribe:
            try: 
                print('Emiting pending events')
                if(self.get_pending_events_on_subscribe and self.join_agent_rooms): 
                    await self._emit_step_events(AgentExecutionStatus.Pending, self.join_agent_rooms)
            except Exception as e:
                print('query-api:: Unable to get pending events', e)            

    async def _subscribe(self):
        if not self.join_account_room and not self.join_agent_rooms:
            raise ValueError('No rooms to join in configuration')
        if not self.socket_client.connected:
            raise ConnectionError('Failed to connect to the WebSocket server.')
        
        async def event_handler(data):
            parsed_data = json.loads(data)
            await self.callback(parsed_data)    

        await self.join_room(self.join_account_room, self.join_agent_rooms)

        if self.subscribe_event_types:
            for event in self.subscribe_event_types:
                print(f"nvm-backend:: Subscribing to event: {event}")
                self.socket_client.on(event, event_handler)
        else:
            self.socket_client.on('step-updated', event_handler)  
            
    async def _emit_step_events(self, status: AgentExecutionStatus = AgentExecutionStatus.Pending, dids: List[str] = []):
        message = { "status": status.value, "dids": dids }
        print(f"nvm-backend:: Emitting step: {json.dumps(message)}")
        await self.socket_client.emit(event='_emit-steps', data=json.dumps(message))

    async def join_room(self, join_account_room: bool, room_ids: Optional[Union[str, List[str]]] = None):
        print(f"event:: Joining rooms: {room_ids} and {self.user_room_id}")

        data = { 'joinAccountRoom': join_account_room }
        
        if room_ids:
            data['joinAgentRooms'] = [room_ids] if isinstance(room_ids, str) else room_ids
        
        await self.socket_client.emit('_join-rooms', json.dumps(data))
        
        print(f"event:: Joined rooms: {room_ids} and {self.user_room_id}")

    async def disconnect(self):
        await self.disconnect_socket()
        print("nvm-backend:: Disconnected from the server")

    def parse_url_to_proxy(self, uri: str) -> str:
        return f"{self.opts.proxy_host}{uri}"
    
    def parse_url_to_backend(self, uri: str) -> str:
        return f"{self.opts.backend_host}{uri}"
    
    def parse_headers(self, additional_headers: dict[str, str]) -> dict[str, str]:
        return {
            **self.opts.headers,
            **additional_headers,
        }

    def get(self, url: str, headers: Optional[Dict[str, str]] = None):
        headers = self.parse_headers(headers or {})

        response = requests.get(url, headers=headers) 
        if response.status_code >= 400:
            raise Exception({"data": response.json(), "status": response.status_code, "headers": response.headers})
        return response


    def post(self, url: str, data: Any, headers: Optional[Dict[str, str]] = None):
        headers = self.parse_headers(headers or {})

        response = requests.post(url, json=data, headers=headers)
        if response.status_code >= 400:
            raise Exception({"data": response.json(), "status": response.status_code, "headers": response.headers})
        return response


    def put(self, url: str, data: Any, headers: Optional[Dict[str, str]] = None):
        headers = self.parse_headers(headers or {})
        
        response = requests.put(url, json=data, headers=headers)
        if response.status_code >= 400:
            raise Exception({"data": response.json(), "status": response.status_code, "headers": response.headers})
        return response
      
    def delete(self, url: str, data: Any, headers: Optional[Dict[str, str]] = None):
        headers = self.parse_headers(headers or {})
        
        response = requests.delete(url, json=data, headers=headers)
        if response.status_code >= 400:
            raise Exception({"data": response.json(), "status": response.status_code, "headers": response.headers})
        return response

    def get_service_token(self, service_did: str) -> ServiceTokenResultDto:
        """
        Gets the service token.

        Args:
            service_did (str): The DID of the service.

        Returns:
            ServiceTokenResultDto: The result of the creation operation.

        Raises:
            HTTPError: If the API call fails.

        Example:
            response = your_instance.get_service_token(service_did="did:nv:xyz789")
            print(response)
        """
        url = f"{self.opts.backend_host}/api/v1/payments/service/token/{service_did}"
        response = self.get(url)
        return ServiceTokenResultDto.model_validate(response.json()['token'])